from typing import Sequence
from clingo.ast import (
    AST,
    ASTType,
)
from clingo.ast import parse_string
from .xclingo_ast import (
    is_choice_rule,
    is_constraint,
    is_disyunctive_head,
    is_theory_head,
)

from ._translator import (
    SupportTranslator,
    AnnotationTranslator,
    RelaxedConstraintTranslator,
    AnnotationNames,
)


def xclingo_annotation(rule_ast):
    if rule_ast.ast_type == ASTType.Rule and rule_ast.head.ast_type == ASTType.TheoryAtom:
        if rule_ast.head.term.name in AnnotationNames.all():
            return rule_ast.head.term.name
    return None


class Preprocessor:
    def __init__(self) -> None:
        self._translation = ""

    def reset(self) -> None:
        self._translation = ""

    def preprocess_rule(self, rule_ast: AST) -> Sequence[AST]:
        raise RuntimeError("This method is intended to be override")

    def _add_to_translation(self, rule_asts: Sequence[AST]):
        """Adds the given rule to the internal translation.

        Args:
            rule_ast (AST): the rule to add to the translation.
        """
        for ra in rule_asts:
            self._translation += f"{ra}\n"

    def _add_comment_to_translation(self, comment: str):
        """Adds a comment to the internal translation.

        Args:
            comment (str): the comment to add to the translation.
        """
        self._translation += f"% {comment}\n"

    def process_program(self, program: str):
        """Translates a given program into its xclingo translation. The translation starts with a comment containing the program name.

        Args:
            program (str): program to be translated.
            name (str, optional): Defaults to "".
        """
        self._translation = ""
        parse_string(
            program,
            lambda ast: self._add_to_translation(self.preprocess_rule(ast)),
        )
        return self._translation


class XClingoAnnotationPreprocessor(Preprocessor):
    def __init__(self) -> None:
        super().__init__()

    def reset(self) -> None:
        super().reset()

    def preprocess_rule(self, rule_ast: AST) -> None:
        yield rule_ast

    def process_program(self, program: str):
        return super().process_program(program.replace("%!", "&"))


class ConstraintRelaxer(Preprocessor):
    """Relaxes the constraints in the program. This is, traced constraints become rules with special
    heads.

    Created heads have the form: xclingo_violated_constraint(<number>). This number ID is
    independent from the number of rules in the program.
    """

    def __init__(self, keep_annotations: bool = False):
        super().__init__()
        self._constraint_count = 1
        self.there_is_a_label = False
        self._keep_annotations = keep_annotations
        self._relaxed_constraint_translator = RelaxedConstraintTranslator()

    def reset(self):
        super().reset()
        self._constraint_count = 1
        self.there_is_a_label = False

    def _increment_constraint_count(self):
        """Returns the current ID for constraints and increment the internal counter."""
        n = self._constraint_count
        self._constraint_count += 1
        return n

    def preprocess_rule(self, rule_ast: AST):
        """Preprocess the given rule and adds the result to the translation.

        Labelled constraints are transformed into their relaxed form. The rest of the program is
        unchanged.

        Args:
            rule_ast (AST): rule to be preprocessed and added to the translation.
        """
        self._add_comment_to_translation(rule_ast)
        # Ignore #shows
        if rule_ast.ast_type == ASTType.ShowSignature:
            return

        if rule_ast.ast_type != ASTType.Rule:
            yield rule_ast  # Things that are not rules are just passed
        else:
            annotation_name = xclingo_annotation(rule_ast)

            if annotation_name is None:  # Rules
                rule_id = self._increment_constraint_count()
                if is_constraint(rule_ast) and self.there_is_a_label is True:
                    for translated_constraint in self._relaxed_constraint_translator.translate(
                        rule_id, rule_ast
                    ):
                        yield translated_constraint
                else:
                    yield rule_ast
                self.there_is_a_label = False

            else:  # Annotations
                if annotation_name == "trace_rule":
                    self.there_is_a_label = True
                if self._keep_annotations:
                    yield rule_ast


class XClingoPreprocessor(Preprocessor):
    """Translates a given program into the xclingo format.

    For every original rule ID is given and the following happens:
    - the original rule is added as a comment in the translation.
    - a support rule (_xclingo_sup/3) is added to the translation. The body of this rule is the positive
    part of the original rule, written in terms of _xclngo_model/1 facts.
    - a fired_body rule (_xclingo_fbody/3) is added to the translation. The body of this rule is the
    positive part of the orignal rule, written in terms of _xclingo_f_atom/1 facts.
    - if the rule is traced, a label rule (_xclingo_label/2) is added to the translation. The body of
    this rule directly depends of the corresponding _xclingo_f/3 fact of the traced rule.
    - traced constraints are also translated into their xclingo forms. Not traced constraints are
    ignored.

    xClingo anotations are translated into its xclingo forms:
    - %!show_trace: translated into a rule with head _xclingo_show_trace/1.
    - %!trace: translated into a rule with head _xclingo_trace/2.
    """

    def __init__(self):
        super().__init__()
        self._rule_count = 1
        self._last_trace_rule = None
        self._annotation_translator = AnnotationTranslator()
        self._support_translator = SupportTranslator()

    def reset(self):
        super().reset()
        self._rule_count = 1
        self._last_trace_rule = None

    def _increment_rule_count(self):
        """Returns the current ID for rules and increment the internal counter."""
        n = self._rule_count
        self._rule_count += 1
        return n

    def translate_annotation(self, annotation_name: str, rule_ast: AST):
        for translated_annotation in self._annotation_translator.translate(
            annotation_name,
            rule_ast,
            self._rule_count,
        ):
            yield translated_annotation

    def translate_rule(
        self,
        rule_id: int,
        disjunction_id: int,
        original_head: AST,
        original_body: Sequence[AST],
    ):
        for translated_rule in self._support_translator.translate(
            rule_id, disjunction_id, original_head, original_body, self._last_trace_rule
        ):
            yield translated_rule

    def preprocess_rule(self, rule_ast: ASTType.Rule):
        """Translates a given rule into its xclingo translation and adds it to the translation.
        Before every addition, a comment containing the original rule is also added.

        Not traced constraints will be ignored.

        Args:
            rule_ast (ASTType.Rule): rule to be translated.
        """
        # Ignore #shows
        if rule_ast.ast_type == ASTType.ShowSignature:
            return
        # TODO: what to do with externals?
        # TODO: which other things
        self._add_comment_to_translation(rule_ast)
        if rule_ast.ast_type != ASTType.Rule:
            yield rule_ast  # Things that are not rules are just passed
        else:
            # Checks if it is an xclingo annotation
            annotation_name = xclingo_annotation(rule_ast)
            if annotation_name is not None:
                if annotation_name == "trace_rule":
                    self._last_trace_rule = rule_ast
                else:
                    for translated_annotation in self.translate_annotation(
                        annotation_name,
                        rule_ast,
                    ):
                        yield translated_annotation
            else:
                rule_id = self._increment_rule_count()

                if is_theory_head(rule_ast):  # Is not an xclingo theory atom
                    yield rule_ast
                    return

                if is_choice_rule(rule_ast):
                    for cond_lit in rule_ast.head.elements:

                        for r in self.translate_rule(
                            rule_id,
                            0,
                            cond_lit.literal,
                            list(cond_lit.condition) + list(rule_ast.body),
                        ):
                            yield r

                elif is_disyunctive_head(rule_ast):
                    disjunction_id = 0
                    for cond_lit in rule_ast.head.elements:
                        for r in self.translate_rule(
                            rule_id,
                            disjunction_id,
                            cond_lit.literal,
                            list(cond_lit.condition) + list(rule_ast.body),
                        ):
                            yield r
                        disjunction_id += 1
                elif not is_constraint(rule_ast):
                    for r in self.translate_rule(rule_id, 0, rule_ast.head, rule_ast.body):
                        yield r

                self._last_trace_rule = None



from ._preprocessor import Preprocessor
from ._preprocessor import XClingoAnnotationPreprocessor
from ._preprocessor import ConstraintRelaxer
from ._preprocessor import XClingoPreprocessor


class PreprocessorPipeline:
    def __init__(self):
        self.preprocessors = list()

    def register_at_beginning(self, preprocessor: Preprocessor):
        self.preprocessors.insert(0, preprocessor)

    def register_at_end(self, preprocessor: Preprocessor):
        self.preprocessors.append(preprocessor)

    def translate(self, name: str, program: str):
        translation = program
        for p in self.preprocessors:
            translation = p.process_program(translation)
        return f"%%%%%%%% {name} %%%%%%%%\n{translation}"


class ConstraintRelaxerPipeline(PreprocessorPipeline):
    def __init__(self):
        super().__init__()
        self.register_at_end(XClingoAnnotationPreprocessor())
        self.register_at_end(ConstraintRelaxer())


class DefaultExplainingPipeline(PreprocessorPipeline):
    def __init__(self):
        super().__init__()
        self.register_at_end(XClingoAnnotationPreprocessor())
        self.register_at_end(XClingoPreprocessor())


class ConstraintExplainingPipeline(PreprocessorPipeline):
    def __init__(self):
        super().__init__()
        self.register_at_end(XClingoAnnotationPreprocessor())
        self.register_at_end(ConstraintRelaxer(keep_annotations=True))
        self.register_at_end(XClingoPreprocessor())



from clingo.core import MessageCode


class XclingoLogger:
    def __init__(self) -> None:
        self._no_labels = None
        self._no_show_trace = None

    def logger(self, _code, msg):
        """Logger TODO: more detail."""
        if _code == MessageCode.AtomUndefined:
            if "xclingo_muted" in msg:
                return
            if "_xclingo_label" in msg:
                self._no_labels = True
                return
            if "_xclingo_show_trace" in msg:
                self._no_show_trace = True
        print(msg)

    def print_messages(self):
        """Prints messages from the logger."""
        if self._no_labels:
            print("xclingo info: any atom or rule has been labelled.")
        if self._no_show_trace:
            print("xclingo info: any atom has been affected by a %!show_trace annotation.")



from clingo import String
from clingo.symbol import SymbolType


class XClingoContext:
    """Xclingo context class."""

    def label(self, text, tup):
        """Given the text of a label and a tuple of symbols, handles the variable instantiation
        and returns the processed text label."""
        if text.type == SymbolType.String:
            text = text.string
        else:
            text = str(text).strip('"')
        for val in tup.arguments:
            text = text.replace("%", val.string if val.type == SymbolType.String else str(val), 1)
        return [String(text)]



from typing import Sequence, Union
from clingo import TheoryTerm
from clingo.ast import (
    AST,
    Location,
    ASTType,
    Sign,
)
from clingo.ast import (
    Rule,
    Variable,
    Function,
)
from ._ast_shortcuts import (
    xclingo_dependency_head_literal,
    xclingo_body_aggregate_literal,
    xclingo_conditional_literal,
    literal,
    collect_free_vars,
    loc,
)

########### Constants ###########

_SHOW_TRACE_HEAD = "_xclingo_show_trace"
_MUTE_HEAD = "_xclingo_muted"
_MUTE_RULE_HEAD = "_xclingo_muted_body"
_TRACE_HEAD = "_xclingo_label"

_MODEL_WRAPPER = "_xclingo_model"
_SUP_HEAD = "_xclingo_sup"
_DEPENDS_HEAD = "_xclingo_depends"
_F_HEAD = "_xclingo_f"
_XCLINGO_CONSTRAINT_HEAD = "_xclingo_violated_constraint"


########### Rule types ###########


class XclingoAST:
    def __init__(self):
        pass

    def get_ast(self):
        return self._ast


class XclingoRule(XclingoAST):
    def __init__(
        self, rule_id: int, disjunction_id: int, location: Location, head: AST, body: Sequence[AST]
    ):
        self._ast = Rule(
            location=loc,
            head=self.translate_head(rule_id, disjunction_id, head, body),
            body=list(self.translate_body(body)),
        )

    def translate_head(self, rule_id: int, disjunction_id: int, head: AST, body: Sequence[AST]):
        raise NotImplementedError("This method is intended to be override")

    def translate_body(self, body: Sequence[AST]):
        raise NotImplementedError("This method is intended to be override")


####### Reference Lit


class ReferenceLit:
    def __init__(
        self,
        func_name: str,
        rule_id: Union[int, AST],
        disjunction_id: Union[int, AST],
        head: AST,
        body: Sequence[AST],
        **kwargs,
    ):
        self._reference_lit = literal(
            func_name,
            [rule_id, disjunction_id, head, list(collect_free_vars(body))],
        )
        super().__init__(
            rule_id=rule_id, disjunction_id=disjunction_id, head=head, body=body, **kwargs
        )


class SupLit(ReferenceLit):
    def __init__(self, **kwargs):
        super().__init__(func_name=_SUP_HEAD, **kwargs)


class FLit(ReferenceLit):
    def __init__(self, **kwargs):
        super().__init__(func_name=_F_HEAD, **kwargs)


####### Bodies


class DoNothingBody:
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def translate_body(self, body: Sequence[AST]):
        for lit in body:
            yield lit


class ModelBody:
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def translate_body(self, body: Sequence[AST]):
        for lit in body:
            if lit.ast_type == ASTType.Literal:
                if lit.atom.ast_type == ASTType.SymbolicAtom:
                    yield literal(_MODEL_WRAPPER, [lit], sign=lit.sign)

                elif lit.atom.ast_type == ASTType.BodyAggregate:
                    yield xclingo_body_aggregate_literal(self.translate_body, lit)

                else:
                    yield lit

            elif lit.ast_type == ASTType.ConditionalLiteral:
                yield xclingo_conditional_literal(_MODEL_WRAPPER, self.translate_body, lit)

            else:
                yield lit


####### Rules


class SupportRule(ModelBody, SupLit, XclingoRule):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def translate_head(self, rule_id: int, disjunction_id: int, head: AST, body: Sequence[AST]):
        return self._reference_lit


class RelaxedConstraint(DoNothingBody, XclingoRule):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def translate_head(self, rule_id: int, disjunction_id: int, head: AST, body: Sequence[AST]):
        return literal(
            _XCLINGO_CONSTRAINT_HEAD, [rule_id, list(collect_free_vars(body))], sign=Sign.NoSign
        )


class DependsRule(ModelBody, SupLit, XclingoRule):
    def __init__(self, extra_body: Sequence[AST], causes: Sequence[AST], **kwargs):
        self._extra_body = extra_body
        self._causes = causes
        super().__init__(**kwargs)

    def translate_head(self, rule_id: int, disjunction_id: int, head: AST, body: Sequence[AST]):
        return xclingo_dependency_head_literal(_DEPENDS_HEAD, self._reference_lit, self._causes)

    def translate_body(self, body: Sequence[AST]):
        for lit in super().translate_body(self._extra_body):
            yield lit
        yield self._reference_lit


####### Annotations that mark atoms
class AutoSafeTheory:
    def __init__(self, head: AST, body: Sequence[AST], **kwargs):
        # We need the name and the arguments of the symbol to be able to create the annotation
        term = head.elements[0].terms[0]
        if term.ast_type == ASTType.SymbolicTerm:
            name, args = term.symbol.name, term.symbol.arguments
        elif term.ast_type == ASTType.TheoryUnparsedTerm:  # strong negation
            args = term.elements[0].term.arguments
            if len(term.elements[0].operators) == 1 and term.elements[0].operators[0] == "-":
                name = "-" + term.elements[0].term.name
            else:
                name = term.elements[0].term.name
        else:
            name, args = term.name, term.arguments

        super().__init__(
            head=head,
            body=[literal(name, args)] + list(body),
            **kwargs,
        )


class MarkAnnotation(AutoSafeTheory, ModelBody, XclingoRule):
    def __init__(self, wrapper: str, **kwargs):
        self._wrapper = wrapper
        super().__init__(
            **kwargs,
        )

    def translate_head(self, rule_id: int, disjunction_id: int, head: AST, body: Sequence[AST]):
        return literal(self._wrapper, [head.elements[0].terms[0]], sign=Sign.NoSign)


class ShowTraceAnnotationRule(MarkAnnotation):
    def __init__(self, **kwargs):
        super().__init__(wrapper=_SHOW_TRACE_HEAD, **kwargs)


class MuteAnnotationRule(MarkAnnotation):
    def __init__(self, **kwargs):
        super().__init__(wrapper=_MUTE_HEAD, **kwargs)


####### MarkRule


class MuteRuleAnnotation(DoNothingBody, XclingoRule):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def translate_head(self, rule_id: int, disjunction_id: int, head: AST, body: Sequence[AST]):
        return literal(_MUTE_RULE_HEAD, [rule_id], sign=Sign.NoSign)


####### Traces


class TraceAnnotation:
    def __init__(self, labelled: AST, label: AST, vars: Sequence[AST], **kwargs):
        self.labelled = labelled
        self.label = label
        self.vars = vars
        super().__init__(**kwargs)

    def translate_head(self, rule_id: int, disjunction_id: int, head: AST, body: Sequence[AST]):
        external_function = Function(
            location=loc,
            name="label",
            arguments=[self.label, Function(loc, "", self.vars, False)],
            external=True,
        )
        return literal(_TRACE_HEAD, [self.labelled, external_function], sign=Sign.NoSign)


class TraceRuleAnnotationRule(TraceAnnotation, FLit, XclingoRule):
    def __init__(self, trace_head: AST, **kwargs):
        super().__init__(
            labelled=Variable(loc, "Head"),
            label=trace_head.elements[0].terms[0],
            vars=trace_head.elements[0].terms[1:],
            disjunction_id=Variable(loc, "DisID"),
            head=Variable(loc, "Head"),
            **kwargs,
        )

    def translate_body(self, body: Sequence[AST]):
        yield self._reference_lit


class TraceAnnotationRule(TraceAnnotation, AutoSafeTheory, ModelBody, XclingoRule):
    def __init__(self, head: AST, **kwargs):
        super().__init__(
            labelled=head.elements[0].terms[0],
            label=head.elements[0].terms[1],
            vars=head.elements[0].terms[2:],
            head=head,
            **kwargs,
        )



from argparse import ArgumentParser, FileType
import sys
from ._version import __version__ as xclingo_version

from pathlib import Path


def check_options():
    """Handles arguments of xclingo"""
    parser = ArgumentParser(
        description="Tool for explaining (and debugging) ASP programs", prog="xclingo"
    )
    parser.add_argument(
        "--version",
        action="version",
        version="xclingo {version}".format(version=xclingo_version),
        help="Prints the version and exists.",
    )
    parser.add_argument(
        "--output",
        type=str,
        choices=[
            "ascii-trees",
            "translation",
            "graph-models",
            "render-graphs",
        ],
        default="ascii-trees",
        help="""Determines the format of the output. "translation" will output the translation 
        together with the xclingo logic program. "graph-models" will output the explanation 
        graphs following clingraph format.""",
    )
    parser.add_argument(
        "--debug-output",
        type=str,
        choices=[
            "none",
            "explainer-program",
            "unsat-solver-program",
            "unsat-explainer-program",
        ],
        default="none",
    )
    parser.add_argument(
        "--outdir",
        type=Path,
        help="Only takes effect when used together with --output=render-graphs.",
    )
    parser.add_argument(
        "--picklefile",
        type=str,
        help="""""",
    )
    parser.add_argument(
        "--auto-tracing",
        type=str,
        choices=["none", "facts", "all"],
        default="none",
        help="""Automatically creates traces for the rules of the program. Default: none.
        - 'facts' will create labels for every fact rule in the program.
        - 'all' will create labels for every rule in the program.""",
    )
    parser.add_argument(
        "--constraint-explaining",
        type=str,
        choices=["minimize", "all"],
        default="minimize",
        help="""Explains traced constraints of the program. Default: unsat.
        - 'unsat' will only explain constraints if original program is UNSAT. In such a case
        the ocurrence of violated constraints will be minimized when solving the original program.
        - 'minimize' acts as 'unsat' but the first UNSAT check is skipped. Directly minimizes the
        constraints when solving the original program. Useful when the original program is known
        UNSAT but it takes long to check it.
        - 'all' will explain all constraints and will not minimize them before explaining.""",
    )
    parser.add_argument(
        "--print-models",
        action="store_true",
        help="Prints each original model before its explanations. ",
    )
    parser.add_argument(
        "-n",
        nargs=2,
        default=(1, 1),
        type=int,
        help="Number of answer sets and number of desired explanations.",
    )
    parser.add_argument(
        "infiles", nargs="+", type=FileType("r"), default=sys.stdin, help="ASP program"
    )
    return parser.parse_known_args()



from typing import Sequence
from clingo.ast import AST, ASTType, Sign, Variable, BooleanConstant
from typing import Callable, Iterator, Sequence, Union
from clingo.ast import (
    AST,
    Location,
    Position,
    ASTType,
    Sign,
)
from clingo.ast import (
    Literal,
    SymbolicAtom,
    Function,
    BodyAggregate,
    BodyAggregateElement,
    ConditionalLiteral,
    SymbolicTerm,
    Pool,
    Variable,
)
from clingo import Number, String


loc = Location(
    Position("", 0, 0),
    Position("", 0, 0),
)

######### Body checks #########


def propagates(lit_list: Sequence[AST]):
    """Captures the part of a body that propagate causes.
    This is, the positive part of the body of a rule. Comparison literals are ignored.

    Args:
        lit_list (Sequence[AST]): list of literals to be processed. Normally
        a rule's body.

    Yields:
        AST: literals that propagate cause.
    """
    for lit in lit_list:
        if (
            lit.ast_type != ASTType.ConditionalLiteral
            and lit.sign == Sign.NoSign
            and lit.atom.ast_type == ASTType.SymbolicAtom
        ):
            yield lit


def aggregates(lit_list: Sequence[AST]):
    """Captures the part of a body that is an aggregate.

    Args:
        lit_list (Sequence[AST]): list of literals to be processed. Normally
        a rule's body.

    Yields:
        AST: literals that are aggregates.
    """
    for lit in lit_list:
        if (
            lit.ast_type != ASTType.ConditionalLiteral
            and lit.sign == Sign.NoSign
            and lit.atom.ast_type == ASTType.BodyAggregate
        ):
            for e in lit.atom.elements:
                yield e


def conditional_literals(lit_list: Sequence[AST]):
    """Captures the part of a body that is a conditional literal.

        Args:
            lit_list (Sequence[AST]): list of literal
    def wrapped_literal(wrapper_name: str, literal: AST):
        return literal(wrapper_name, [literal.atom.symbol], literal.sign)
    s to be processed. Normally
            a rule's body.

        Yields:
            AST: literals that are conditional literals.
    """
    for lit in lit_list:
        if lit.ast_type == ASTType.ConditionalLiteral:
            yield lit


def collect_free_vars(lit_list: Sequence[AST]):
    def collect_vars(lit: AST):
        if lit.atom.symbol.ast_type == ASTType.UnaryOperation:
            arguments = lit.atom.symbol.argument.arguments
        else:
            arguments = lit.atom.symbol.arguments
        for arg in arguments:
            if arg.ast_type == ASTType.Variable:
                yield str(arg.name)

    seen_vars, unsafe_vars = set(), set()
    for lit in lit_list:
        # handle conditional literals
        if lit.ast_type == ASTType.ConditionalLiteral:
            for var_name in collect_vars(lit.literal):
                unsafe_vars.add(var_name)
            continue

        # handle comparisons
        if lit.atom.ast_type == ASTType.Comparison:
            if lit.atom.left.ast_type == ASTType.Variable:
                seen_vars.add(str(lit.atom.left.name))
            if lit.atom.right.ast_type == ASTType.Variable:
                seen_vars.add(str(lit.atom.right.name))

        if lit.atom.ast_type == ASTType.BodyAggregate:
            if (
                lit.atom.left_guard is not None
                and lit.atom.left_guard.term.ast_type == ASTType.Variable
            ):
                seen_vars.add(str(lit.atom.left_guard.term.name))
            if (
                lit.atom.right_guard is not None
                and lit.atom.right_guard.term.ast_type == ASTType.Variable
            ):
                seen_vars.add(str(lit.atom.right_guard.term.name))

        # Skip negative literals
        elif lit.sign != Sign.NoSign:
            continue

        # handle positive body literals
        elif lit.atom.ast_type == ASTType.SymbolicAtom:
            for var_name in collect_vars(lit):
                seen_vars.add(var_name)
            continue

    for var_name in seen_vars:
        if var_name not in unsafe_vars:
            yield Variable(loc, var_name)


######### Check type of rule #########


def is_constraint(rule_ast):
    if rule_ast.ast_type == ASTType.Rule:
        if hasattr(rule_ast.head, "atom"):
            return (
                rule_ast.head.atom.ast_type == ASTType.BooleanConstant
                and rule_ast.head.atom == BooleanConstant(0)
            )
    return False


def is_choice_rule(rule_ast):
    return (
        rule_ast.ast_type == ASTType.Rule
        and rule_ast.head.ast_type == ASTType.Aggregate
        and hasattr(rule_ast.head, "function") == False
    )


def is_disyunctive_head(rule_ast):
    return rule_ast.head.ast_type == ASTType.Disjunction


def is_theory_head(rule_ast):
    return rule_ast.head.ast_type == ASTType.TheoryAtom


######### Shortcuts for creating ASTs #########


def handle_type(element: Union[AST, str, int, Sequence]):
    if isinstance(element, AST):
        if element.ast_type == ASTType.Literal:
            element = element.atom
        return element
    elif isinstance(element, str):
        return SymbolicTerm(loc, String(element))
    elif isinstance(element, int):
        return SymbolicTerm(loc, Number(element))
    elif hasattr(element, "__len__") and hasattr(element, "__getitem__"):
        return Function(loc, "", list([handle_type(item) for item in element]), False)
    else:
        raise ValueError("Unknown type: {}".format(type(element)))


def wrap_symbols(wrapper_name: str, symbols: Sequence):
    return SymbolicAtom(Function(loc, wrapper_name, [handle_type(item) for item in symbols], False))


def literal(func_name: str, args: Sequence, sign: Sign = Sign.NoSign):
    return Literal(
        loc,
        sign,
        wrap_symbols(func_name, args),
    )


def xclingo_dependency_head_literal(function_name: str, refernce_lit: AST, causes: Sequence[AST]):
    return literal(func_name=function_name, args=[refernce_lit, Pool(loc, causes)])


def xclingo_body_aggregate_literal(
    transformer_function: Callable[[Sequence[AST]], Iterator[AST]], lit: AST
):
    return Literal(
        loc,
        lit.sign,
        BodyAggregate(
            loc,
            left_guard=lit.atom.left_guard,
            function=lit.atom.function,
            elements=[
                BodyAggregateElement(
                    terms=list(transformer_function(e.terms)),
                    condition=list(transformer_function(e.condition)),
                )
                for e in lit.atom.elements
            ],
            right_guard=lit.atom.right_guard,
        ),
    )


def xclingo_conditional_literal(
    lit_wrapper: str,
    transformer_function: Callable[[Sequence[AST]], Iterator[AST]],
    conditional_literal: AST,
):
    return ConditionalLiteral(
        loc,
        literal=literal(
            lit_wrapper, [conditional_literal.literal], sign=conditional_literal.literal.sign
        ),
        condition=list(transformer_function(conditional_literal.condition)),
    )



try:
    from importlib.resources import read_text
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    from importlib_resources import read_text


def load_xclingo_extension(extension_filename: str):
    return read_text(__package__, resource=extension_filename)



from ._version import __version__ as xclingo_version


def print_version():
    print(f"xclingo version {xclingo_version}")


def print_header(args):
    print_version()
    print(f"Reading from {' '.join([f.name for f in args.infiles])}")



from clingo.ast import AST, Location, Position
from typing import Sequence

from .xclingo_ast import (
    ShowTraceAnnotationRule,
    TraceAnnotationRule,
    MuteAnnotationRule,
    MuteRuleAnnotation,
    SupportRule,
    DependsRule,
    TraceRuleAnnotationRule,
    RelaxedConstraint,
)
from .xclingo_ast import (
    propagates,
    aggregates,
    conditional_literals,
)
from enum import Enum

# TODO: fix location
loc = Location(
    Position("", 0, 0),
    Position("", 0, 0),
)


class RelaxedConstraintTranslator:
    def __init__(self) -> None:
        pass

    def translate(self, rule_id: int, constraint_ast: AST):
        yield RelaxedConstraint(
            location=None,
            disjunction_id=None,
            rule_id=rule_id,
            head=constraint_ast.head,
            body=constraint_ast.body,
        ).get_ast()


class AnnotationNames(dict):
    _SHOW_TRACE = "show_trace"
    _TRACE = "trace"
    _TRACE_RULE = "trace_rule"
    _MUTE = "mute"
    _MUTE_BODY = "mute_body"

    @staticmethod
    def all():
        return [
            v for k, v in AnnotationNames.__dict__.items() if not k.startswith("__") and k != "all"
        ]


class AnnotationTranslator:
    def __init__(self):
        self._annotation_asts = {
            AnnotationNames._SHOW_TRACE: ShowTraceAnnotationRule,
            AnnotationNames._TRACE: TraceAnnotationRule,
            AnnotationNames._MUTE: MuteAnnotationRule,
            AnnotationNames._MUTE_BODY: MuteRuleAnnotation,
        }

    def translate(self, annotation_name: str, rule_ast: AST, rule_id: int):
        if annotation_name in self._annotation_asts:
            yield self._annotation_asts[annotation_name](
                location=None,
                head=rule_ast.head,
                body=rule_ast.body,
                rule_id=rule_id,
                disjunction_id=None,
            ).get_ast()
        else:
            raise RuntimeError(f"Unknown annotation name: {annotation_name}")


class RuleTranslator:
    def __init__(self, rule, depends, trace_rule) -> None:
        self._rule = rule
        self._depends = depends
        self._trace_rule = trace_rule

    def translate(
        self, rule_id, disjunction_id, rule_head: AST, rule_body: Sequence[AST], trace_rule_ast: AST
    ) -> Sequence[AST]:
        yield self._rule(
            rule_id=rule_id,
            disjunction_id=disjunction_id,
            location=None,
            head=rule_head,
            body=rule_body,
        ).get_ast()
        if rule_body:
            causes = list(propagates(rule_body))
            if causes:
                yield self._depends(
                    rule_id=rule_id,
                    disjunction_id=disjunction_id,
                    location=None,
                    head=rule_head,
                    body=rule_body,
                    extra_body=[],
                    causes=causes,
                ).get_ast()
        for agg_element in aggregates(rule_body):
            causes = list(propagates(agg_element.condition))
            if causes:
                yield self._depends(
                    rule_id=rule_id,
                    disjunction_id=disjunction_id,
                    location=None,
                    head=rule_head,
                    body=rule_body,
                    extra_body=agg_element.condition,
                    causes=causes,
                ).get_ast()
        for cond_lit in conditional_literals(rule_body):
            causes = list(propagates(cond_lit.condition))
            yield self._depends(
                rule_id=rule_id,
                disjunction_id=disjunction_id,
                location=None,
                head=rule_head,
                body=rule_body,
                extra_body=cond_lit.condition,
                causes=[cond_lit.literal] + causes,
            ).get_ast()
        if self._trace_rule is not None and trace_rule_ast is not None:
            yield self._trace_rule(
                rule_id=rule_id,
                location=None,
                body=rule_body,
                trace_head=trace_rule_ast.head,
            ).get_ast()


class SupportTranslator(RuleTranslator):
    def __init__(self) -> None:
        super().__init__(SupportRule, DependsRule, TraceRuleAnnotationRule)



class ExplanationControlError(RuntimeError):
    pass


class ExplanationControlParsingError(ExplanationControlError):
    pass


class ExplanationControlGroundingError(ExplanationControlError):
    pass



from typing import Sequence, TextIO
from xclingo import XclingoControl
from xclingo.preprocessor import (
    DefaultExplainingPipeline,
    ConstraintRelaxerPipeline,
    ConstraintExplainingPipeline,
)
from xclingo.extensions import load_xclingo_extension
from xclingo.error import ModelControlGroundingError, ModelControlParsingError
from xclingo.explainer.error import ExplanationControlParsingError, ExplanationControlGroundingError
from ._utils import print_header, print_version
from ._arguments_handler import check_options

from clingraph.orm import Factbase
from clingraph import compute_graphs as compute_clingraphs
from clingraph.graphviz import render


def read_files(files: Sequence[TextIO]):
    return "\n".join([file.read() for file in files])


def _init_xclingo_control(
    args,
    unknown_args,
    program,
    constraints=False,
):
    xclingo_control = XclingoControl(
        [str(args.n[0])] + unknown_args,
        n_explanations=str(args.n[1]),
        solving_preprocessor_pipeline=ConstraintRelaxerPipeline() if constraints else None,
        explaining_preprocessor_pipeline=ConstraintExplainingPipeline() if constraints else None,
    )

    if args.auto_tracing != "none":
        xclingo_control.extend_explainer(
            "base", [], load_xclingo_extension(f"autotrace_{args.auto_tracing}.lp")
        )

    if args.output == "render-graphs":
        xclingo_control.extend_explainer("base", [], load_xclingo_extension("graph_locals.lp"))
        xclingo_control.extend_explainer("base", [], load_xclingo_extension("graph_styles.lp"))

    if constraints:
        xclingo_control.extend_explainer(
            "base", [], load_xclingo_extension("violated_constraints_show_trace.lp")
        )
        if args.constraint_explaining == "minimize":
            xclingo_control.extend_explainer(
                "base", [], load_xclingo_extension("violated_constraints_minimize.lp")
            )

    xclingo_control.add("base", [], program)
    xclingo_control.ground([("base", [])])

    return xclingo_control


def print_explainer_program(args):
    from xclingo.xclingo_lp import FIRED_LP, GRAPH_LP, SHOW_LP

    print_version()
    print(FIRED_LP)
    print(GRAPH_LP)
    print(SHOW_LP)

    if args.auto_tracing != "none":
        print(load_xclingo_extension(f"autotrace_{args.auto_tracing}.lp"))

    if args.output == "render-graphs":
        print(load_xclingo_extension("graph_locals.lp"))
        print(load_xclingo_extension("graph_styles.lp"))

    if args.debug_output == "explainer-program":
        pipe = DefaultExplainingPipeline()
    elif args.debug_output == "unsat-explainer-program":
        pipe = ConstraintExplainingPipeline()
    elif args.debug_output == "unsat-solver-program":
        pipe = ConstraintRelaxerPipeline()

    if len(args.infiles) > 0:
        print(pipe.translate("translation", read_files(args.infiles)))


def render_graphs(args, xclingo_control: XclingoControl):
    nmodel = 0
    for x_model in xclingo_control.solve():
        nmodel += 1
        nexpl = 0
        for graph_model in x_model.explain_model():
            nexpl += 1
            render(
                compute_clingraphs(
                    Factbase.from_model(
                        graph_model, prefix="_xclingo_", default_graph="explanation"
                    ),
                    graphviz_type="digraph",
                ),
                name_format=f"explanation{nmodel}-{nexpl}" + "_{graph_name}",
                format="png",
                directory=args.outdir if args.outdir else "out/",
            )

    if nmodel > 0:
        print("Images saved in ./out/")

    return nmodel == 0


def solve_explain(args, xclingo_control: XclingoControl):
    nmodel = 0
    for x_model in xclingo_control.solve():
        nmodel += 1
        print(f"Answer: {nmodel}")
        if args.print_models:
            print(x_model)
        nexpl = 0
        for graph_model in x_model.explain_model():
            nexpl += 1
            print(f"##Explanation: {nmodel}.{nexpl}")
            if args.output == "graph-models":
                print(graph_model)
            else:
                for sym in graph_model.show_trace:
                    e = graph_model.explain(sym)
                    if e is not None:
                        print(e)
        print(f"##Total Explanations:\t{nexpl}")
    if nmodel > 0:
        print(f"Models:\t{nmodel}")
        return False
    else:
        return True


def into_pickle(args, xclingo_control: XclingoControl, save_on_unsat=False):
    buf = []
    for x_model in xclingo_control.solve():
        for graph_model in x_model.explain_model():
            buf.append(frozenset(str(s) for s in graph_model.symbols(shown=True)))

    if len(buf) == 0 and not save_on_unsat:
        return True

    import pickle

    with open(args.picklefile, "wb") as picklefile:
        pickle.dump(frozenset(buf), picklefile)
        print(f"Results saved as frozen sets at {args.picklefile}")

    return False


def ground_solve_explain(args, unknown_args, programs):

    xclingo_control = _init_xclingo_control(args, unknown_args, programs)
    if args.picklefile:  # default value: ""
        unsat = into_pickle(args, xclingo_control, save_on_unsat=False)
    elif args.output == "render-graphs":
        unsat = render_graphs(args, xclingo_control)
    else:
        unsat = solve_explain(args, xclingo_control)

    xclingo_control = _init_xclingo_control(args, unknown_args, programs, constraints=True)
    if unsat:
        if args.picklefile:
            into_pickle(args, xclingo_control, save_on_unsat=True)
        elif args.output == "render-graphs":
            render_graphs(args, xclingo_control)
        else:
            print("UNSATISFIABLE")
            print(f"Relaxing constraints... (mode={args.constraint_explaining})")
            solve_explain(args, xclingo_control)


def main():
    """Main function. Checks command line arguments and acts in consequence."""
    args, unknown_args = check_options()

    # Prints translation and explainer lp and exits
    if args.debug_output != "none":
        print_explainer_program(args)
        return 0

    # Prints translation and exits
    if args.output == "translation":
        print(DefaultExplainingPipeline().translate("translation", read_files(args.infiles)))
        return 0

    print_header(args)
    programs = read_files(args.infiles)

    try:
        ground_solve_explain(args, unknown_args, programs)
    except ModelControlParsingError as e:
        print("*** ERROR: (clingo, original program)", e)
        exit(1)
    except ModelControlGroundingError as e:
        print("*** ERROR: (clingo, original program)", e)
        exit(1)
    except ExplanationControlParsingError as e:
        print("*** ERROR: (xclingo, explainer program)", e)
    except ExplanationControlGroundingError as e:
        print("*** ERROR: (xclingo, explainer program)", e)


if __name__ == "__main__":
    main()



from typing import Iterator, Sequence, Union
from clingo import Logger, Model, Function
from clingo.control import Control
from xclingo.explanation import ExplanationGraphModel
from xclingo.xclingo_lp import FIRED_LP, GRAPH_LP, SHOW_LP
from ._utils import XClingoContext
from .error import ExplanationControlGroundingError, ExplanationControlParsingError


class Explainer:
    """Xclingo explainer class. Obtains the solutions from an xclingo program and turns them into explanations."""

    def __init__(
        self,
        arguments: Sequence[str] = [],
        logger: Union[Logger, None] = None,
        message_limit: int = 20,
    ):
        self.arguments = arguments
        self.logger = logger
        self.message_limit = message_limit

        self.programs = []

    def add(self, name, parameters, program):
        self.programs.append((name, parameters, program))

    def _compute_graphs(
        self, model: Model, on_explanation=None, context=None
    ) -> Iterator[ExplanationGraphModel]:
        """Return the ExplanationGraphModel instances for the given (original's program) model.

        Args:
            model (Model): original's program model.
            context (_type_, optional): Context class for the internal Control instance. Defaults to None.

        Yields:
            ExplanationGraphModel: each explanation for the model.
        """
        # Initializes control
        ctl = Control(
            arguments=self.arguments + ["--project=project"],
            logger=self.logger,
            message_limit=self.message_limit,
        )
        # Adds xclingo program
        ctl.add("base", [], FIRED_LP)
        ctl.add("base", [], GRAPH_LP)
        ctl.add("base", [], SHOW_LP)
        # Translations and extensions
        try:
            for name, parameters, program in self.programs:
                ctl.add(name, parameters, program)
        except RuntimeError as e:
            raise ExplanationControlParsingError(e)
        # Adds model (TODO: is it possible to do this with just one control?)
        with ctl.backend() as back:
            for sym in model.symbols(atoms=True):
                back.add_rule([back.add_atom(Function("_xclingo_model", [sym], True))], [], False)
        # Ground
        try:
            ctl.ground([("base", [])], context=context if context is not None else XClingoContext())
        except RuntimeError as e:
            raise ExplanationControlGroundingError(e)
        # Solve
        if on_explanation is not None:
            ctl.solve(on_model=on_explanation)
        else:
            with ctl.solve(yield_=True) as it:
                for graph_model in it:
                    yield ExplanationGraphModel(graph_model)



from clingo import Symbol, Model
from collections import defaultdict


class ExplanationGraphModel(Model):
    def __init__(self, graph_model: Model):
        super().__init__(graph_model._rep)

    @property
    def show_trace(self):
        if not hasattr(self, "_index"):
            self.index
        return self._show_trace

    @property
    def index(self):
        if not hasattr(self, "_index"):
            # Builds graph from model
            setattr(self, "_show_trace", [])
            setattr(self, "_index", defaultdict())
            for s in self.symbols(shown=True):
                if len(s.arguments) == 1:  # _xclingo_show_trace(Atom)
                    self._show_trace.append(s.arguments[0])
                elif len(s.arguments) == 2:  # _xclingo_edge((Caused, Cause), explanation)
                    if s.name == "_xclingo_edge":
                        caused, cause = s.arguments[0].arguments
                    else:  # _xclingo_link(ToExplainAtom, Cause)
                        caused, cause = s.arguments
                    self._index.setdefault(caused, Explanation(caused))
                    self._index.setdefault(cause, Explanation(cause))
                    self._index[caused].add_cause(self._index[cause])
                else:  # _xclingo_attr(node, Atom, label, Label)
                    self._index.setdefault(s.arguments[1], Explanation(s.arguments[1]))
                    self._index[s.arguments[1]].add_label(str(s.arguments[3]))
        return self._index

    def explain(self, symbol: Symbol):
        return self._index.get(symbol, None)


class Explanation:
    def __init__(self, atom: Symbol, labels: set = None, causes: set = None):
        self.labels = labels
        if self.labels is not None:
            if not isinstance(labels, set):
                raise TypeError('"labels" arguments must be of type set.')
        else:
            self.labels = set()

        self.causes = causes
        if self.causes is not None:
            if not isinstance(causes, list):
                raise TypeError('"child_nodes" arguments must be of type list.')
        else:
            self.causes = list()

        self.atom = atom

    def preorder_iterator(self, only_labelled=False):
        if only_labelled and not self.labels:
            stack = [iter(self.causes)]
        else:
            stack = [iter([self])]
        level = 1
        while stack:
            try:
                current = next(stack[-1])
                yield (current, level)
                level += 1
                stack.append(iter(current.causes))
            except StopIteration:
                stack.pop()
                level += -1

    def ascii_tree(self):
        def ascii_branch(level):
            if level > 0:
                return "  |" * (level) + "__"
            else:
                return ""

        expl = "  *\n"
        for node, level in self.preorder_iterator(only_labelled=True):
            expl += "{branch}{text}\n".format(
                branch=ascii_branch(level),
                text=node.get_node_text(),
            )
        return expl

    def get_node_text(self):
        return ";".join(sorted(list(self.labels)))

    def add_cause(self, cause):
        self.causes.append(cause)

    def add_label(self, label):
        self.labels.add(label)

    def __str__(self):
        return self.ascii_tree()



from clingo import Symbol
from typing import Sequence, Union, Tuple, Any
from clingo import Model, Logger
from clingo.control import Control
from xclingo.explanation import ExplanationGraphModel
from xclingo.explainer import Explainer
from xclingo.explainer._utils import XClingoContext

from xclingo.preprocessor import PreprocessorPipeline
from xclingo.preprocessor import DefaultExplainingPipeline

from xclingo.explainer._logger import XclingoLogger

from xclingo.error import ModelControlGroundingError, ModelControlParsingError
from xclingo.explainer.error import ExplanationControlGroundingError, ExplanationControlParsingError


class XClingoModel(Model):
    def __init__(self, original_model: Model, explainer: Explainer):
        super().__init__(original_model._rep)
        self._explainer = explainer

    def explain_model(self) -> Sequence[ExplanationGraphModel]:
        return self._explainer._compute_graphs(self, context=XClingoContext())


class XclingoControl(Control):
    """Interface to xclingo pipeline. Grounding, solving, explaining."""

    def __init__(
        self,
        arguments: Sequence[str] = [],
        logger: Union[Logger, None] = None,
        message_limit: int = 20,
        n_explanations: str = "1",
        solving_preprocessor_pipeline: PreprocessorPipeline = None,
        explaining_preprocessor_pipeline: PreprocessorPipeline = None,
    ):
        # Solver control
        super().__init__(arguments=arguments, logger=logger, message_limit=message_limit)
        self.pre_solving_pipeline = (
            PreprocessorPipeline()
            if solving_preprocessor_pipeline is None
            else solving_preprocessor_pipeline
        )

        # Explainer control
        self.pre_explaining_pipeline = (
            DefaultExplainingPipeline()
            if explaining_preprocessor_pipeline is None
            else explaining_preprocessor_pipeline
        )
        self.expl_logger = XclingoLogger()
        constants = []
        for i, a in enumerate(arguments):
            if a in ("-c", "--const"):
                constants.extend(arguments[i : i + 2])
        self.explainer = Explainer([n_explanations] + constants, logger=self.expl_logger.logger)

    def add(self, name: str, parameters: Sequence[str], program: str) -> None:
        """It adds a program to the control.
        Args:
            name (str): name of program block to add.
            program (str): a logic program in ASP format.
        """
        try:
            super().add(name, parameters, self.pre_solving_pipeline.translate(name, program))
        except RuntimeError as e:
            raise ModelControlParsingError(e)
        try:
            self.explainer.add(
                name, parameters, self.pre_explaining_pipeline.translate(name, program)
            )
        except RuntimeError as e:
            raise ExplanationControlParsingError(e)

    def ground(self, parts: Sequence[Tuple[str, Sequence[Symbol]]], context: Any = None) -> None:
        try:
            super().ground(parts, context)
        except RuntimeError as e:
            raise ModelControlGroundingError(e)

    def extend_explainer(self, name: str, parameters: Sequence[str], program: str) -> None:
        self.explainer.add(name, parameters, program)

    def add_show_trace(self, atom: Symbol, conditions: Sequence[Tuple[bool, Symbol]] = []):
        self.explainer.add("base", [], f"_xclingo_show_trace({str(atom)}) :- .")

    def solve(self, on_unsat=None) -> Sequence[XClingoModel]:
        """Returns a generator of xclingo.explanation.Explanation objects. If on_explanation is not None, it is called for each explanation."""
        if on_unsat:
            super().solve(on_unsat=on_unsat)
        with super().solve(yield_=True) as solution_iterator:
            for model in solution_iterator:
                yield XClingoModel(model, self.explainer)

    def _default_output(self) -> None:
        output = ""
        nanswer = 0
        for xModel in self.solve():
            nanswer += 1
            output += f"Answer: {nanswer}\n"
            output += " ".join(map(str, xModel.symbols(shown=True))) + "\n"
            nexpl = 0
            for graphModel in xModel.explain_model():
                nexpl += 1
                output += f"##Explanation: {nanswer}.{nexpl}\n"
                output += "\n".join(
                    [graphModel.explain(s).ascii_tree() for s in graphModel.show_trace]
                )
                output += "\n"
            output += f"##Total Explanations:\t{nexpl}\n"
        output += f"Models:\t{nanswer}\n"
        return output



# Xclingo ASTs
from ._xclingo_ast import (
    ShowTraceAnnotationRule,
    TraceAnnotationRule,
    MuteAnnotationRule,
    MuteRuleAnnotation,
    SupportRule,
    DependsRule,
    TraceRuleAnnotationRule,
    RelaxedConstraint,
)

# Shortcuts for checking things
from ._ast_shortcuts import is_choice_rule, is_constraint, is_disyunctive_head, is_theory_head

# Shortcuts for inspecting bodies
from ._ast_shortcuts import (
    propagates,
    aggregates,
    conditional_literals,
)



try:
    from importlib.resources import read_text
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    from importlib_resources import read_text

FIRED_LP = read_text(__package__, "xclingo_fired.lp")
GRAPH_LP = read_text(__package__, "xclingo_graph.lp")
SHOW_LP = read_text(__package__, "xclingo_show.lp")
